"""
Extend an object by adding an attribute or method.
"""
from slugify import slugify

from . import objverify
from .objprope import get_class


def validate_attribute_name(name):
	""" Modify a string to become a valid/legal lower case attribute """
	name = slugify(name).replace('-', '_')
	return name


def add_attribute(cls, attr_name, attr_value):
	""" Add a valid attribute name to a class and set the attribute value """
	attr_name = validate_attribute_name(attr_name)
	setattr(cls, attr_name, attr_value)


#  Code solution from: http://stackoverflow.com/questions/20078816/replace-non-ascii-characters-with-a-single-space
def remove_non_ascii(text, replacement=' '):
	""" 
	When mining for data, you may come across special characters
	that will raise a 'UnicodeEncodeError', this may include printing, 
	storing in a database or other reasons.
	
	Call the function, passing the inflicted string, to replace those characters 
	with, by default a space. Instead of a space the second paramater, which is
	optional allows you to specify the replacement character.
	""" 
	return str(re.sub( r'[^\x00-\x7F]+', replacement, text ))




def make_get_has_header_methods_for_obj_with_response(obj):
	""" 
	The passed object must have the attribute `response`, and that response
	must be an object that has an attribute named `headers` that is a dict.

	Dynamically create methods for the headers in the response object, 
	and append those methods to the class of the object (not the instance).
	"""

	objverify.has_response(obj)
	objverify.has_headers(obj.response)
	if not hasattr(cls, 'response'):
		raise AttributeError('The instance of %r has no attribute `response`' % cls) 
	elif not hasattr(cls.response, 'headers'): 
		raise AttributeError('%r instance\'s `response` attribute has no attribute `headers`' % cls) 

	#### Methods Wrapper ##########################################
	def _make_method(header_name, create_has_method=True):
		""" Create a method for a header """			
		#### Methods to make #######################
		def _has_header(self):
			""" Return Flase if header value is None """
			if header_name not in self.response.headers:
				self.response.headers[header_name] = None
			return self.response.headers[header_name] != None

		def _get_header(self): 
			""" Returns the header value """
			if header_name not in self.response.headers:
				self.response.headers[header_name] = None
			return self.response.headers[header_name]
		############################################
	
		# Create a valid function name
		func_name = header_name
		for INVALID in ('-', ' '):
			func_name = func_name.replace(INVALID, '_')

		if create_has_method:
			_method = _has_header
			_method.__name__ = 'has_%s_header' %  func_name
		else:
			_method = _get_header
			_method.__name__ = 'get_%s_header' % func_name
		return _method
	###############################################################


	obj_cls = get_class(obj)
	for header_name in self.response.headers:
		_has_method = _make_method(header_name, create_has_method=True)
		_get_method = _make_method(header_name, create_has_method=False)
		setattr(obj_cls, _has_method.__name__, _has_method)
		setattr(obj_cls, _get_method.__name__, _get_method)






import requests
from bs4 import BeautifulSoup as BS

from .urls import LazyUrl
from .objprope import get_class, get_class_name


__doc__ = \
""" 
Wrapping the isinstance method, verify an objects type.
If an object is not of the specified type raise a TypeError.
Ignoring the error is possible by providing a True value to the optional
ignore parameter in the function calls.  
"""


def is_type_of(obj, cls, ignore=False):
	""" Checks if an object is/of a specified class """
	if isinstance(obj, cls):
		return True
	elif ignore: 
		return False
	raise TypeError('%r is not of type %r' % (get_class(obj), cls))


def has_attr_of_type(obj, attrname, cls, ignore=False):
	""" 
	Checks if an object has a specified attribute, and checks if the
	attribute is of a specified type.
	"""
	type_error = False
	if hasattr(obj, attrname):
		attr = getattr(obj, attrname)
		if is_type_of(cls, attr, ignore=True): 
			return True
		type_error = True
	
	elif ignore: 
		return False
	elif type_error:
		raise TypeError('%s is of type %r, expected type %r' %\
			('%s.%s' % (get_class_name(obj), attrname), get_class(attr), cls))
	raise AttributeError('%r has no attribute `%s`' % (get_class(obj), attrname))




def is_dict(obj, ignore=False):
	""" Checks if an object is type dict """
	return is_type_of(obj, dict, ignore)

def is_lazy_url(obj, ignore=False):
	""" Check if an object is type LazyUrl """
	return is_type_of(obj, LazyUrl, ignore)

def is_soup(obj, ignore=False):
	""" Checks if an object is type BeautifulSoup """
	return is_type_of(obj, BS, ignore)

def has_soup(obj, ignore=False):
	return has_attr_of_type(obj, 'soup', BS, ignore)


def is_response(obj, ignore=False):
	""" Checks if an object is of type requests.Response """
	return is_type_of(obj, requests.Response, ignore)

def has_response(obj, ignore=False):
	return has_attr_of_type(obj, 'response', requests.Response, ignore)


def has_headers(obj, ignore=False):
	return has_attr_of_type(obj, 'headers', dict, ignore)






def get_class(obj): 
	return obj.__class__

def get_class_name(obj): 
	return obj.__class__.__name__



def get_methods_on_class(cls):
	""" 
	Returns a list of tuples containing of all the method names and methods in a class 
	i.e [(method_name1, method1), (method_name2, method2), ...]
	"""
	methods = []
	methods_to_ignore = object.__dict__.keys() # __init__, __str__, etc.

	for attr_name, attr_value in cls.__dict__.items():
		match_str = '<function %s at' % attr_name

		if match_str in str(attr_value):
			if attr_name not in methods_to_ignore:
				methods += [(attr_name, attr_value)]
	return methods







import urlparse


### URL API ####################################################################
class LazyUrl(object):
	""" 
	A more friendly API to urls, using the urlparse library to split and join
	urls.
	"""
	### Variable Declaration ##################################
	ALLOWED_SCHEMES = ('http', 'https')
	scheme, host, path, params, query, fragment = ('','','','','','')
	###########################################################

	### Builtin Methods #######################################
	def __init__(self, url):
		self.set_url(url)	

	def __repr__(self): return '<LazyUrl: %s>' % self.url
	def __str__(self): return self.url
	def __unicode__(self): return unicode(self.url)
	def __iter__(self):
		for p in (self.scheme, self.host, self.path, 
		self.params, self.query, self.fragment,):
			yield p

	def __eq__(self, other):
		if other == self.url:
			return self.url == True
		if isinstance(other, self.__class__):
			return other.__dict__ == self.__dict__
		return False
	def __ne__(self, other):
		return not self.__eq__(other)
	###########################################################

	### Modualarzed Methods ###################################
	#### urlparser Wrappers ############
	def join_url(self, join=True):
		""" Iterate over the url pieces and join them together. """
		if join:
			self.url = urlparse.urlunparse(self)
	####################################


	### Full Path Methods ##############
	def get_full_path(self):
		""" Return the url without the scheme and host parts """
		idx = len(self.scheme) + len(self.host) + 3 # +3 for ://
		return self.url[idx:]

	def clear_full_path(self):
		""" Remove all url pieces except the scheme and host """
		self.set_path('', False)
		self.set_params('', False)
		self.set_query('', False)
		self.set_fragment('', False)
	####################################


	### Set Methods ####################
	def set_url(self, url):
		""" Replace the url, check if it's valid, attempt a fix  """
		self.url = url
		self.parse_url()
		self._fix_broken_url()

	def set_scheme(self, scheme, join=True):
		""" Add/Replace the url scheme """
		self.scheme = scheme
		self.join_url(join)

	def set_host(self, host, join=True):
		""" Add/Replace the url host """
		self.host = host
		self.join_url(join)

	def set_path(self, path, join=True):
		""" Add/Replace the url path """
		self.path = path
		self.join_url(join)

	def set_params(self, params, join=True):
		""" Add/Replace the url params """
		self.params = params
		self.join_url(join)

	def set_query(self, query, join=True):
		""" Add/Replace the url query """
		self.query = query
		self.join_url(join)

	def set_fragment(self, fragment, join=True):
		""" Add/Replace the url fragment """
		self.fragment = fragment
		self.join_url(join)
	####################################
	###########################################################


	### Prope / Inspection Mehods #############################
	def print_overview(self, url=None):
		if isinstance(url, urlparse.ParseResult):
			url.host = url.netloc
			url.url = url.geturl()
		else:
			url = self
	
		print '\nOverview:\n'
		print '\tUrl: %s\n' % url.url
		print '\tScheme: %s' % url.scheme
		print '\tHost: %s' % url.host
		print '\tPath: %s' % url.path	
		print '\tQuery: %s' % url.query
		print '\tFragment: %s\n' % url.fragment
	###########################################################



	### Rules / Correction Methods #############################
	def parse_url(self):
		""" 
		Delegate the urlpieces from the returned urlpare.urlparse() method which is a urlparse.ParseResult()
		object. The reason being that the ParseResult() object inherits from a collections.namedtuple() object
		which restricts access to attributes so that modification/set operations become impossible. 
		Also rename the `netloc` attribute to the more appropiate name, `host`.
		"""
		self.scheme, self.host, self.path, self.params,\
		self.query, self.fragment = urlparse.urlparse(self.url)

	def _fix_broken_url(self, output_results=False):
		""" Locate any bugs in the url, and attempt to fix them. """
		url = urlparse.urlparse(self.url)

		if url.scheme not in self.ALLOWED_SCHEMES:
			self.set_scheme(self.ALLOWED_SCHEMES[0], join=False)

		if url.netloc == '':
			# The host got placed in the path, with no path, move it
			if '/' not in url.path:
				self.set_host(url.path)
				self.set_path('')
			# The host got placed with path, seperate them
			else:
				path = url.path.split('/')
				if path[0] == '' and len(path) > 1:
					path.pop(0)
				host = path.pop(0)
				path = '/'.join(path)
				
				self.set_host(host, join=False)
				self.set_path(path)

		if output_results:
			self.print_overview(url)

	###########################################################
################################################################################







from .objprope import get_class, get_class_name, get_methods_on_class

from .objverify import is_type_of, is_dict, is_lazy_url, is_soup, is_response
from .objverify import has_attr_of_type, has_soup, has_response, has_headers

from .objextend import validate_attribute_name, add_attribute
from .objextend import make_get_has_header_methods_for_obj_with_response

from .urls import LazyUrl



from setuptools import setup


setup(
	name='crispys_webkit',
	version='0.1.12',

	author ='Brandon Nadeau',
	author_email='brandonmnadeau@hotmail.com',

	packages=['crispys_webkit'],

	url='http://github.com/crispycret/crispys_webkit',

	license='MIT',

	description='Some common web functionalites that I use.',
	#long_description=open('README.txt').read(),

	install_requires = [
		'requests',
		'beautifulsoup4',
		'python-slugify',
	],

	zip_safe=False,
)
